/*****************************************************************
*  Author: Pere Urb�n Bayes                                      *
*  e-mail: purbon@ccd.uab.es                                     *
*  Description :  Source file for the main core functions.       *          
*  File Name: ubc.c                                              *
*  Last Changes  : 10-06-2005                                    *
******************************************************************/
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "io.h"
#include "large.h"
#include "small.h"  
#include "ubc.h" 


 
/*******************************************************************************
 * FunctionName: large_kernel
 * Author: Pere Urb�n Bayes 
 * e-mail: purbon@ccd.uab.es
 * Description: This function do the main work to calc the kernel set and the
 *              leader set for the largest codes.
 * Input:
 *   (Value)  File: A pointer to the code data source file. 
 *   (Value)  wordLength : The code word length.    
 *   (Value)  codeCardinal : The code cardinal.
 *   (Reference) leaderCardinal : The leader set cardinal.  
 *   (Reference) kernelCardinal : The kernel set cardinal. 
 *   (Reference) kernelBound : The kernel Bound, used to speed the our work. 
 * Output: 
 *    large_wCodeWord array if all goes good, but if the method found any
 *    problem it will return NULL. 
 ******************************************************************************/ 
 
large_wCodeWord * large_kernel(FILE * File, int wordLength, int codeCardinal, int * leaderCardinal, int * kernelCardinal,int * kernelBound) 
{
  /* Some auxiliari variables, index, etc.. */   
       
  int ePosition;
  int i,j,k,l;
 
  large_codeWord word;  
    
  int root1, root2;
  int count;
   
  /* Structure of Data to easy calc the kernel and the leaders. */
  large_wCodeWord * codeElements;
   
  /* Temporary array to easy and fast calc the Kernel Elements. */
  large_codeWord * kernelSet; 
  
  /* Temporary array to make easy and fast to calc the leader Set. */
  large_codeWord * leaderSet;
  
  /* Code word's loading from the source file. */
    
  codeElements = (large_wCodeWord *) DecimalLoad(File,wordLength, codeCardinal);
 
//  /* Dinamic memory alloc to use the temporary and dinamic arrays. */
//  
//  /* Code Word */ 
  if ((word = (large_codeWord)calloc(limbs,sizeof(small_codeWord))) == NULL) 
  {
  	  printf("Problems with memory allocation in kernel array\n.");
   	  return MEM_ERROR;
  }
  
  /* Kernel Set */ 
  if ((kernelSet= (large_codeWord*)calloc(codeCardinal,sizeof(large_codeWord))) == NULL) 
  {
  	  printf("Problems with memory allocation in kernel array\n.");
   	  return MEM_ERROR;
  }
  /* Leader Set */
   
  if ((leaderSet = (large_codeWord*)calloc(codeCardinal,sizeof(large_codeWord))) == NULL) 
  {
  	  printf("Problems with memory allocation in Lider Array\n.");
   	  return MEM_ERROR;
  }    
  
  /* Kernel and leader set code word's array */
  for ( i = 0; i < codeCardinal; i++) 
  {
   	/* Kernel Set */
   	if ((kernelSet[i] = (large_codeWord)calloc(limbs,sizeof(small_codeWord))) == NULL) 
   	{
       printf("Problems with memory allocation in kernel array\n.");
       return MEM_ERROR;
    }
    
    /* Leader Set */
  	if ((leaderSet[i] = (large_codeWord)calloc(limbs,sizeof(small_codeWord))) == NULL) 
   	{
       printf("Problems with memory allocation in Lider Array element\n.");
       return MEM_ERROR;
    }
     
  } 
      
  /* End of the dinamic memory alloc. Here start the real algorithm */
  
  /* We sort the codeElements to let us use eficient binary search */
  large_sortq(codeElements,codeCardinal);
 	

  /***********************************************************************
   * Since we supous that the element zero is always on the kernel set,  *
   * firstly we insert it to the set.                                    *                            
   ***********************************************************************/
     
  codeElements[0].state = KERNEL_ELEMENT;
  for(k = 0; k < limbs; k++)  kernelSet[0][k] = 0;   
  *kernelCardinal = 1;
  
  /***********************************************************************
   * To try to make the calc speddy, if the codeCardinal is a power of   *
   * two, whe can supous to have a Kernel set bound to codeCardinal / 4. *
   * In the other hand, the kernelBound will be the codeCardinal.        *
   *                                                                     *
   * Technicaly :                                                        *
   *                                                                     *
   *      if | Kernel | =  | C | / 2                                     *
   *                                                                     *
   *         then C = Kernel U Kernel + x                                *
   *                                                                     *
   *       ==> C is linear.                                              *
   *                                                                     *
   *      But if C is a lienar code, then Kernel = C.                    *
   *                                                                     *
   ***********************************************************************/
   
  if (codeWordWeight(codeCardinal) == 1) 
     *kernelBound = ( codeCardinal >> 2 );
   else
     *kernelBound = codeCardinal;
   
  i = 1; *leaderCardinal = 0; count = 1;
  
  /*************************************************************************
   * Now we can proced to calc the kernel and the leader set's.            *
   *************************************************************************/
        
  while ( (count<codeCardinal) && ( *kernelCardinal <= *kernelBound ) )
  {    
     /* If the current word is not processed yet */
     
     if (codeElements[i].state == NOT_PROCESSED) 
     {
       j = 0; ePosition = 0;	   

  	   /*******************************************************************
  	    * Trying to verify if the current word belongs to the kernel Set. *	         
  	    * If we do C + w ( for all the word into the code) and the result *
  	    * is a code word into the code, then w belongs to the kernel Set. *
  	    *******************************************************************/       
  		    
	   while ((j < codeCardinal)&&(ePosition != NOT_FOUND)) 
  	   {        
          /* binary add, bite to bite */
          for(k = 0; k < limbs; k++)
                word[k] = codeElements[i].codeWord[k] ^ codeElements[j].codeWord[k]; 
                            
		  ePosition = large_BinarySearch(word,codeElements,codeCardinal);
		  j++;
	   }  
	 
	   if (ePosition != NOT_FOUND) 
	   {  
	    /* This element is into the kernel. */

        /*********************************************************************
         * Now we ara going to calc the transfered kernel elements.          *
         *                                                                   *
         * If we know that w belongs to the kernel, we know that w plus      *
         * all the word that are know into the kernel, belongs to the kernel.*
         * In the other case ( w doesnt belongs to the kernel ), we          *
         * are in the same case, all the plus calculus arent into the kernel.*        
         *********************************************************************/
       
		 for (j=0;j<*kernelCardinal;j++) 
		 { 
         	for(k = 0; k < limbs; k++)
	      	  	  word[k] = codeElements[i].codeWord[k]^kernelSet[j][k];
	            		     	  	  
		    ePosition = large_BinarySearch(word,codeElements,codeCardinal);
		    codeElements[ePosition].state = KERNEL_ELEMENT;
			 
		    for(k = 0; k < limbs; k++)
                  kernelSet[(*kernelCardinal)+j][k] = word[k];
          
            count++;
   	                 
     	   /*****************************************************************
     		* At this point we know that the current code word plus all the *
     		* leader set elements are not in to the kernel.                 *
     		*                                                               *
     		* Here the leader sets will be recalculated, because the last   *
     		* operation that we comented, generate a new word that belongs  *
     		* to some coset leader, that if this word was not processed yet *
     		* it only have to be added to the him coset, but if this word   *
     		* was yet processed we have to join the two coset leader because*
     		* are the same coset.                                           *
     		*****************************************************************/ 
     		       
		    for(l=0;l<*leaderCardinal;l++)
		    {
		     	for(k = 0; k < limbs; k++)
                      word[k] = codeElements[ePosition].codeWord[k]^leaderSet[l][k];     		        

                root1 = large_BinarySearch(leaderSet[l],codeElements,codeCardinal);
		        root2 = large_BinarySearch(word,codeElements,codeCardinal);
		     	         
		        if (codeElements[root2].state == NOT_PROCESSED ) 
		        {   
		        	/* The word must be joined to the him coset. */ 
     		        large_Union(leaderSet,leaderCardinal,codeElements,root1,root2);                         
	                count++;   	     	           
		        }      
        	    else
                {   
                	/* Join the two cosets leaders because are the seame. */                           	
                    root2 = large_Find(codeElements,root2); 
     		 	    large_Union(leaderSet,leaderCardinal,codeElements,root1,root2);                           	
		        }
		     	        
     	    }
     	 }   
   	   *kernelCardinal <<= 1;
   	   }
       else  
	   {
	   	/********************************************************************
	 	 * This word doesent belongs to the kernel.                         *
	 	 * It means that this word plus all the kernel set will generate a  *
	 	 * new coset leader.                                                *
	 	 ********************************************************************/	      
	         	
	      codeElements[i].state = LEADER;
	      count++;
	      
	      for(k = 0; k < limbs; k++)
               	leaderSet[*leaderCardinal][k] = codeElements[i].codeWord[k];  	
          (*leaderCardinal)++;
	      
	      root1 = i; 
	         	
	      for(j=1;j<*kernelCardinal;j++)
	      {  	
             for(k = 0; k < limbs; k++)
                   word[k] = codeElements[i].codeWord[k]^kernelSet[j][k];
              		         
  			 root2 = large_BinarySearch(word,codeElements,codeCardinal);                                                           			     		    
             large_Union(leaderSet,leaderCardinal,codeElements,root1,root2);
  			 count++;  			         
          }                 
	   }	           
     }
    i++;
  }

  /* We are going to free the dynamic arrays */
      
  for ( i = 0; i < codeCardinal; i++) 
  {
  	  free(leaderSet[i]);
  	  free(kernelSet[i]); 
  }  
  
      free(kernelSet); 
  	  free(leaderSet);
      free(word); 
        
  return codeElements;
}

int running(char * fileName)
{

  FILE * File;
 	
  small_wCodeWord  * smallCode;
  large_wCodeWord  * largeCode;
 	
  int wordLength,codeCardinal;
  int leaderCardinal,kernelCardinal,kernelBound;

  int i;

  /*********************************************************** 
   * Open the our code data input file and read some         *
   * important variables to know how to do the correct calc. *
   ***********************************************************/ 

  filePath = fileName;
  	
  binaryInput = false;
    
  if (( File = fopen(filePath, "r+")) == NULL )
  {
     printf("This file %s can not be opened\n",filePath);
     return NOT_OK;      
  }
   
  if (binaryInput == true) fscanf(File,"%d\n",&base); 
        
  fscanf(File, "%d\n",&wordLength);
  fscanf(File, "%d\n",&codeCardinal);
  fscanf(File, "%d\n",&limbs);

  if ( wordLength <= UnsignedIntLength ) 
  {    
  	   /* We are working with an small code */
       smallCode = small_kernel(File,wordLength,codeCardinal,&leaderCardinal,&kernelCardinal,&kernelBound);  
       small_DecimalDataOutput(smallCode,codeCardinal,kernelCardinal,leaderCardinal,kernelBound);
       free(smallCode);
  }
  else
  {    
  	   /* Our code is on the large side. */
       largeCode = large_kernel(File,wordLength,codeCardinal,&leaderCardinal,&kernelCardinal,&kernelBound); 
       large_DecimalDataOutput(largeCode, codeCardinal,kernelCardinal,leaderCardinal, kernelBound);
               
 		for(i=0;i<codeCardinal;i++)
		{
    		free(largeCode[i].codeWord);
 		}
		free(largeCode);
  }
  
  return OK;
}


large_wCodeWord * runLargeCode(char * fileName)
{

  FILE * File;
  large_wCodeWord  * largeCode;
 	
  int wordLength,codeCardinal;
  int leaderCardinal,kernelCardinal,kernelBound;

  int i;

  /*********************************************************** 
   * Open the our code data input file and read some         *
   * important variables to know how to do the correct calc. *
   ***********************************************************/ 

  filePath = fileName;
  	
  binaryInput = false;
    
  if (( File = fopen(filePath, "r+")) == NULL )
  {
     printf("This file %s can not be opened\n",filePath);
     return NOT_OK;      
  }

  fscanf(File, "%d\n",&wordLength);
  fscanf(File, "%d\n",&codeCardinal);
  fscanf(File, "%d\n",&limbs);

   /* Our code is on the large side. */
   largeCode = large_kernel(File,wordLength,codeCardinal,&leaderCardinal,&kernelCardinal,&kernelBound); 

 return largeCode;
 
}
